/*
  MDXplayer :  YM2151 emulator access routines

  Made by Daisuke Nagano <breeze.nagano@nifty.ne.jp>
  Apr.16.1999
 */
#include "app.h"///

/* ------------------------------------------------------------------ */

///#ifdef HAVE_CONFIG_H
///# include "config.h"
///#endif /* HAVE_CONFIG_H */

///#include <stdio.h>

///#include "version.h"
///#include "mdx.h"
///#include "mdx2151.h"
///#include "ym2151.h"
///#include "pcm8.h"

/* ------------------------------------------------------------------ */

typedef struct _YM2151_LFO {

  int flag;

  int form;
  int clock;
  int depth;
  int depth2;

} YM2151_LFO;

typedef struct _DETUNE_VAL {

  int octave;
  int scale;
  int detune;

} DETUNE_VAL;

typedef struct _OPM_WORK {

  int        hlfo_flag;
  int        hlfo_val;
  int        lfo_delay;
  YM2151_LFO plfo;
  YM2151_LFO alfo;

  long       portament;
  DETUNE_VAL detune;
  int        note;
  int        note_on;
  int        volume;
  int        pan;
  int        total_level[4];
  int        step;
  int        freq_reg[3];
  int        algorithm;
  int        slot_mask;

} OPM_WORK;

/* ------------------------------------------------------------------ */

static int ym2151_note[] ={
  0,1,2,4,5,6,8,9,10,12,13,14
};

static int is_vol_set[8][4]={
  {0,0,0,1},
  {0,0,0,1},
  {0,0,0,1},
  {0,0,0,1},
  {0,0,1,1},
  {0,1,1,1},
  {0,1,1,1},
  {1,1,1,1}
};

static int master_vol_table[128] = {
    0, 18, 28, 36, 42, 46, 51, 54, 57, 60, 62, 65, 67, 69, 70, 72,
   74, 75, 77, 78, 79, 81, 82, 83, 84, 85, 86, 87, 88, 89, 90, 90,
   91, 92, 93, 93, 94, 95, 96, 96, 97, 97, 98, 99, 99,100,100,101,
  102,102,103,103,104,104,105,105,105,106,106,107,107,108,108,109,
  109,109,110,110,111,111,111,112,112,112,113,113,113,114,114,114,
  115,115,115,116,116,116,117,117,117,117,118,118,118,119,119,119,
  119,120,120,120,120,121,121,121,122,122,122,122,122,123,123,123,
  123,124,124,124,124,125,125,125,125,125,126,126,126,126,127,127
};

static int ym2151_register_map[256];

static YM2151_LFO hlfo;
static int        hlfo_sync;
static OPM_WORK   opm[MDX_MAX_FM_TRACKS];

///static int ym2151_enable      = FLAG_FALSE;
static int master_volume      = 127;

/* ------------------------------------------------------------------ */
/* local functions */

static void freq_write( int );
static void volume_write( int );

static void reg_write( int, int );
static int  reg_read( int );

/* ------------------------------------------------------------------ */

void ym2151_reg_init(/** MDX_DATA *mdx **/) {

  int i;

///  if ( mdx->is_use_fm == FLAG_FALSE ) {
///    ym2151_enable = FLAG_FALSE;
///    return;
///  }

  /* Init OPM Emulator */
  /*
    number of OPM = 1
    clock         = 4MHz
    sample rate   = PCM8_MASTER_PCM_RATE ( typically 44.1 kHz )
    sample bits   = 16 bit
    */

///  YM2151Init( 1, 4000*1000, mdx->dsp_speed );
  YM2151ResetChip(/** 0 **/);

///  ym2151_enable = FLAG_TRUE;
  master_volume = 127;

  for ( i=0 ; i<8 ; i++ ) {
    reg_write( 0x08, 0*8 + i );    /* KON */
  }
  reg_write( 0x0f, 0 );            /* NE, NFREQ */
  reg_write( 0x18, 0 );            /* LFRQ */
  reg_write( 0x19, 0*128 + 0 );    /* AMD */
  reg_write( 0x19, 1*128 + 0 );    /* AMD */
  reg_write( 0x1b, 0*64  + 0 );    /* CT, W */

  for ( i=0 ; i<8 ; i++ ) {
    reg_write( 0x20+i, 3*64 + 0*8 +0 ); /* LR, FL, CON */
    reg_write( 0x28+i, 0*16 + 0 );      /* OCT, NOTE */
    reg_write( 0x30+i, 0 );             /* KF */
    reg_write( 0x38+i, 0*16 + 0 );      /* PMS, AMS */
  }
  for ( i=0 ; i<0x20 ; i++ ) {
    reg_write( 0x40+i, 0*16 + 0 );      /* DT1, MUL */
    reg_write( 0x60+i, 0 );             /* TL */
    reg_write( 0x80+i, 0*64 + 0 );      /* KS, AR */
    reg_write( 0xa0+i, 0*128 + 0 );     /* AMS, D1R */
    reg_write( 0xc0+i, 0*64 + 0 );      /* DT2, D2R */
    reg_write( 0xe0+i, 0*16 + 0 );      /* D1L, RR */
  }

  for ( i=0 ; i<MDX_MAX_FM_TRACKS ; i++ ) {

    OPM_WORK *o = &opm[i];

    o->note          = 0;
    o->note_on       = 0;
    o->step          = 0;
    o->pan           = 3;
    o->detune.octave = 0;
    o->detune.scale  = 0;
    o->detune.detune = 0;
    o->portament     = 0;
    o->freq_reg[0]   =-1;
    o->freq_reg[1]   =-1;
    o->freq_reg[2]   =-1;
    o->algorithm     = 0;
    o->volume        = 0;
    o->lfo_delay     = 0;

    o->hlfo_flag     = FLAG_FALSE;
    o->hlfo_val      = 0;

    o->plfo.flag     = FLAG_FALSE;
    o->plfo.form     = 0;
    o->plfo.clock    = 0;
    o->plfo.depth    = 0;

    o->alfo.flag     = FLAG_FALSE;
    o->alfo.form     = 0;
    o->alfo.clock    = 0;
    o->alfo.depth    = 0;

    o->total_level[0]= 0;
    o->total_level[1]= 0;
    o->total_level[2]= 0;
    o->total_level[3]= 127;
  }

  hlfo_sync=FLAG_FALSE;

  hlfo.form   = 0;
  hlfo.clock  = 0;
  hlfo.depth  = 0;
  hlfo.depth2 = 0;

  return;
}

void ym2151_all_note_off( void ) {

  int i,j;

  for ( i=0 ; i<MDX_MAX_FM_TRACKS ; i++ ) {
    ym2151_note_off(i);
  }
  for ( i=0 ; i<MDX_MAX_FM_TRACKS ; i++ ) {
    reg_write( 0x08, 0+i );          /* KON */

    j=reg_read( 0x20+i );
    reg_write( 0x20+i, j&0x3f );     /* LR, FL, CON */
  }

  return;
}

void ym2151_note_on( int track, int n ) {

  OPM_WORK *o = &opm[track];

  if ( o->step != 0 ) {
    o->portament = 0;
  }
  if ( o->note_on == 0 ) { /* for software-LFO w/ tie */
    o->step        = 0;
  }
  o->note        = n;
  o->note_on     = 1;
  o->freq_reg[0] = -1;
  o->freq_reg[1] = -1;
  o->freq_reg[2] = -1;

  reg_write( 0x1b, hlfo.form );
  reg_write( 0x18, hlfo.clock );
  reg_write( 0x19, hlfo.depth );
  reg_write( 0x19, hlfo.depth2 );

  if ( o->hlfo_flag == FLAG_TRUE && hlfo_sync == FLAG_TRUE ) {
    reg_write( 0x01, 0x02 ); /* LFO SYNC */
    reg_write( 0x01, 0x00 );
  }

  freq_write(track);
  volume_write(track);

  return;
}

void ym2151_note_off(int track) {

  opm[track].note_on   = 0;
  opm[track].portament = 0;

  return;
}

void ym2151_set_pan( int track, int val ) {

  if ( val < 0 ) { val = 0; }
  if ( val > 3 ) { val = 3; }

  if ( opm[track].pan != val ) {
    int i,v;
    i = reg_read( 0x20 + track );
    v = i&0x3f;
    reg_write( 0x20 + track, v+(val<<6) );  /* LR */
  }
  opm[track].pan = val;

  return;
}

void ym2151_set_volume( int track, int val ) {

  if ( val < 0 ) { val = 0; }
  if ( val > 127 ) { val = 127; }
  opm[track].volume = val;

  return;
}

void ym2151_set_master_volume( int val ) {

  if ( val < 0 ) { val = 0; }
  if ( val > 127 ) { val = 127; }
  master_volume = master_vol_table[val];

  return;
}

void ym2151_set_detune( int track, int val ) {

  opm[track].detune.octave = val/(64*12);
  opm[track].detune.scale  = (val/64)%12;
  opm[track].detune.detune = val%64;

  return;
}

void ym2151_set_portament( int track, int val ) {

  opm[track].portament = val;
  opm[track].step      = 0;

  return;
}

void ym2151_set_noise_freq( int val ) {

  reg_write( 0x0f, val );

  return;
}

void ym2151_set_lfo_delay( int track, int delay ) {

  opm[track].lfo_delay = delay;

  return;
}

void ym2151_set_plfo( int track, int flag, int form, int clock, int depth ) {

  opm[track].plfo.flag  = flag;
  opm[track].plfo.form  = form;
  opm[track].plfo.clock = clock;
  opm[track].plfo.depth = depth;

  opm[track].step = 0;

  return;
}

void ym2151_set_alfo( int track, int flag, int form, int clock, int depth ) {

  opm[track].alfo.flag  = flag;
  opm[track].alfo.form  = form;
  opm[track].alfo.clock = clock;
  opm[track].alfo.depth = depth;

  return;
}

void ym2151_set_hlfo( int track, int v1, int v2, int v3, int v4, int v5 ) {

  hlfo.form   = v1&0x03;
  hlfo.clock  = v2;
  hlfo.depth  = v3|0x80;
  hlfo.depth2 = v4&0x7f;

  reg_write( 0x1b, hlfo.form   ); /* WAVE FORM */
  reg_write( 0x18, hlfo.clock  ); /* LFRQ */
  reg_write( 0x19, hlfo.depth  ); /* PMD */
  reg_write( 0x19, hlfo.depth2 ); /* AMD */

  if ( (v1&0x40) == 0 ) {
    hlfo_sync = FLAG_FALSE;
  } else {
    hlfo_sync = FLAG_TRUE;
  }

  opm[track].hlfo_flag = FLAG_TRUE;
  opm[track].hlfo_val  = v5;
  reg_write( 0x38 + track, opm[track].hlfo_val );   /* PMS, AMS */

  return;
}

void ym2151_set_hlfo_onoff( int track, int flag ) {

  opm[track].hlfo_flag = flag;

  if ( opm[track].hlfo_flag == FLAG_TRUE ) {
    reg_write( 0x38 + track, opm[track].hlfo_val );   /* PMS, AMS */
  } else {
    reg_write( 0x38 + track, 0 );
  }

  return;
}

void ym2151_set_voice( int track, VOICE_DATA *v ) {

  OPM_WORK *o = &opm[track];
  int i,j,r;

  j = reg_read( 0x20+track );                               /* LR, FL, CON */
  reg_write( 0x20+track, (j&0xc0) + v->v0 );
  o->algorithm = v->con;
  o->slot_mask = v->slot_mask;

  for ( i=0 ; i<4 ; i++ ) {
    r = track + i*8;

    reg_write( 0x40+r, v->v1[i] );    /* DT1, MUL */
    reg_write( 0x80+r, v->v3[i] );    /* KS, AR */
    reg_write( 0xa0+r, v->v4[i] );    /* AME, D1R */
    reg_write( 0xc0+r, v->v5[i] );    /* DT2, D2R */
    reg_write( 0xe0+r, v->v6[i] );    /* SL, RR */

    o->total_level[i] = 127 - v->v2[i];
    if ( is_vol_set[o->algorithm][i] == 0 )
      reg_write( 0x60+r, v->v2[i]&0x7f );   /* TL */
    else
      reg_write( 0x60+r, 127 );             /* TL */
  }

  return;
}

void ym2151_set_reg( int adr, int val ) {

  if ( adr < 0 || adr > 0xff ) { return; }

  if ( val < 0 ) { val = 0; }
  if ( val > 0xff ) { val = 0xff; }

  reg_write( adr, val );

  /* Total level */

  if ( adr >= 0x60 && adr < 0x80 ) {
    if ( val >=0 && val < 128 ) {
      opm[(adr-0x60)/4].total_level[(adr-0x60)%4] = 127-val;
    }
  }

  /* panpot */

  if ( adr >= 0x20 && adr < 0x28 ) {
    opm[adr-0x20].pan = (val&0xc0)>>6;
  }

  return;
}

/* ------------------------------------------------------------------ */

void ym2151_set_freq_volume( int track ) {

  opm[track].step++;
  freq_write( track );
  volume_write( track );

  return;
}

static void freq_write( int track ) {

  OPM_WORK *o = &opm[track];
  int oct, scale, kf;
  int ofs_o, ofs_s, ofs_f;
  long long f;
  int c,d;
  int key;

  /* detune jobs */

  ofs_o = o->detune.octave;  /* octave */
  ofs_s = o->detune.scale;   /* scale */
  ofs_f = o->detune.detune;  /* detune */
    
  /* portament jobs */
    
  if ( o->portament != 0 ) {
    f = o->portament * o->step / 256;
    
    ofs_f += f%64;          /* detune */
    c=0;
    while ( ofs_f > 63 ) { ofs_f-=64; c++; }
    while ( ofs_f <  0 ) { ofs_f+=64; c--; }
    
    ofs_s += c+(f/64)%12;   /* scale */
    c=0;
    while ( ofs_s > 11 ) { ofs_s-=12; c++; }
    while ( ofs_s <  0 ) { ofs_s+=12; c--; }
    
    ofs_o += c+f/(64*12);   /* octave */
  }
    
  /* soft-lfo jobs */

  d = o->step - o->lfo_delay;
  if ( o->plfo.flag == FLAG_TRUE && d >= 0 ) {
    int cl;
    int c;
    switch(o->plfo.form) {
    case 0:
    case 4:
      cl = d % o->plfo.clock;
      f  = o->plfo.depth * cl;
      break;

    case 1:
    case 3:
    case 5:
    case 7:
      cl = d % (o->plfo.clock*2);
      if ( cl<o->plfo.clock ) {
	f =  o->plfo.depth/2;
      } else {
	f = -(o->plfo.depth/2);
      }
      break;

    case 2:
    case 6:
      cl = d % (o->plfo.clock*2);
      c = cl % (o->plfo.clock/2);
      switch ( cl/(o->plfo.clock/2) ) {
      case 0:
	f = o->plfo.depth * c;
	break;
      case 1:
	f = o->plfo.depth * (o->plfo.clock/2 - c );
	break;
      case 2:
	f = -(o->plfo.depth * c);
	break;
      case 3:
	f = -(o->plfo.depth * (o->plfo.clock/2 - c));
	break;

      default:
	f = 0;
	break;
      }
      break;

    default:
      f=0;
      break;
    }

    f/=256;
    ofs_f += f%64;          /* detune */
    c=0;
    while ( ofs_f > 63 ) { ofs_f-=64; c++; }
    while ( ofs_f <  0 ) { ofs_f+=64; c--; }
    
    ofs_s += c+(f/64)%12;   /* scale */
    c=0;
    while ( ofs_s > 11 ) { ofs_s-=12; c++; }
    while ( ofs_s <  0 ) { ofs_s+=12; c--; }
    
    ofs_o += c+f/(64*12);   /* octave */
  }
  
  /* generate f-number */

  kf = ofs_f;
  c=0;
  while ( kf > 63 ) { kf -= 64; c++; }
  while ( kf <  0 ) { kf += 64; c--; }

  scale = c + (o->note%12) + ofs_s;
  c=0;
  while ( scale > 11 ) { scale-=12; c++; }
  while ( scale <  0 ) { scale+=12; c--; }

  scale = ym2151_note[scale];

  oct = c + (o->note/12) + ofs_o;
  if ( oct>7 ) { oct = 7; }
  else if ( oct<0 ) { oct = 0; }


  /* write to register */

  if ( o->note_on != 0 ) { key=o->slot_mask<<3; }
  else { key=0; }


  if ( o->freq_reg[0] != oct*16+scale ) {
    o->freq_reg[0] = oct*16+scale;
    reg_write( 0x28 + track, o->freq_reg[0] );  /* OCT, NOTE */
  }

  if ( o->freq_reg[1] != kf*4 ) {
    o->freq_reg[1] = kf*4;
    reg_write( 0x30 + track, o->freq_reg[1] );  /* KF */
  }

  if ( o->freq_reg[2] != key+track ) {
    o->freq_reg[2] = key+track;
    reg_write( 0x08, o->freq_reg[2] );          /* KEY ON */
  }

  return;
}

static void volume_write( int track ) {

  OPM_WORK *o = &opm[track];
  int i,r;
  int vol;

  /* set volume */

  for ( i=0 ; i<4 ; i++ ) {
    r = track + i*8;
    if ( is_vol_set[o->algorithm][i]==0 ) continue;
    
    vol = master_volume * o->volume * o->total_level[i] / 127 / 127;
    if ( vol > 127 ) vol = 127;
    if ( vol < 0 ) vol = 0;
    vol = 127-vol;
    
    reg_write( 0x60+r, vol );                  /* TL */
  }

  return;
}

/* ------------------------------------------------------------------ */

/* register actions */

static void reg_write( int adr, int val ) {

  if ( adr > 0x0ff ) { return; }
  ym2151_register_map[adr] = val;

///  if ( ym2151_enable == FLAG_TRUE ) {
    YM2151WriteReg( /**0,**/ adr, val );
///  }

  return;
}

static int reg_read( int adr ) {

  if ( adr > 0xff ) { return 0; }
  return ym2151_register_map[adr];
}
