/*
**
** File: ym2151.h - header file for software implementation of YM2151
**                                            FM Operator Type-M(OPM)
**
** (c) 1997,1998,1999,2000 Jarek Burczynski (s0246@priv4.onet.pl)
** Some of the optimizing ideas by Tatsuyuki Satoh
**
** Version 2.140 May, 19th 2000
**
**
** I would like to thank the following people for making this project possible:
**
** Beauty Planets - for making a lot of real YM2151 samples and providing
** additional informations about the chip. Also for the time spent making
** the samples and the speed of replying to my endless requests.
**
** Shigeharu Isoda - for general help, for taking time to scan his YM2151
** Japanese Manual first of all, and answering MANY of my questions.
**
** Nao - for giving me some info about YM2151 and pointing me to Shigeharu.
** Also for creating fmemu (which I stil use to test the emulator).
**
** Aaron Giles and Chris Hardy - they made some samples of one of my favourite
** arcade games so I could compare it to my emulator.
**
** Bryan McPhail and Tim (powerjaw) - for making some samples.
**
** Ishmair - for the datasheet and motivation.
*/

#ifndef _H_YM2151_
#define _H_YM2151_

typedef short INT16;
typedef int UINT32;
typedef UINT32 offs_t;
typedef UINT32 data_t;

/* 16- and 8-bit samples (signed) are supported*/
///#define SAMPLE_BITS 16

///#if (SAMPLE_BITS==16)
	typedef INT16 SAMP;
///#endif
///#if (SAMPLE_BITS==8)
///	typedef signed char SAMP;
///#endif

typedef void (*mem_write_handler)(offs_t offset,data_t data);

/*
** Initialize YM2151 emulator(s).
**
** 'num' is the number of virtual YM2151's to allocate
** 'clock' is the chip clock in Hz
** 'rate' is sampling rate
*/
int YM2151Init(/**int num,**/ int clock, int rate);

/* shutdown the YM2151 emulators*/
///void YM2151Shutdown(void);

/* reset all chip registers for YM2151 number 'num'*/
void YM2151ResetChip(/**int num**/);

/*
** Generate samples for one of the YM2151's
**
** 'num' is the number of virtual YM2151
** '**buffers' is table of pointers to the buffers: left and right
** 'length' is the number of samples that should be generated
*/
void YM2151UpdateOne(/**int num,**/ INT16 /** * **/*buffers, int length);

/* write 'v' to register 'r' on YM2151 chip number 'n'*/
void YM2151WriteReg(/**int n,**/ int r, int v);

/* read status register on YM2151 chip number 'n'*/
///int YM2151ReadStatus(/**int n**/);

/* set interrupt handler on YM2151 chip number 'n'*/
///void YM2151SetIrqHandler(/**int n,**/ void (*handler)(int irq));

/* set port write handler on YM2151 chip number 'n'*/
///void YM2151SetPortWriteHandler(/**int n,**/ mem_write_handler handler);

/////////////////////////////////////////////////////////////////////////////

/*operator data*/
typedef struct{
	unsigned int phase;		/*accumulated operator phase*/
	unsigned int freq;		/*operator frequency*/
	signed   int DT1v;		/*operator DT1 phase inc/decrement*/

	unsigned int MUL;		/*phase multiply*/
	unsigned int DT1;		/*DT1 * 32      */
	unsigned int DT2;		/*DT2 index     */

	signed   int *connect;	/*operator output 'direction'*/

/*Begin of channel specific data*/
/*note: each operator number 0 contains channel specific data*/
	unsigned int FeedBack;	/*feedback shift value for operators 0 in each channel*/
	signed   int FB;		/*operator self feedback value used only by operators 0*/
	signed   int FB0;		/*previous output value*/
	unsigned int KC;		/*operator KC (copied to all operators)*/
	unsigned int KCindex;	/*speedup*/
	unsigned int PMS;		/*channel PMS*/
	unsigned int AMS;		/*channel AMS*/
/*End of channel specific data*/

	unsigned int AMSmask;	/*LFO AMS enable mask*/

	unsigned int state;		/*Envelope state: 4-attack(AR) 3-decay(D1R) 2-sustain(D2R) 1-release(RR) 0-off*/
	unsigned int delta_AR;	/*volume delta for attack phase*/
	unsigned int TL;		/*Total attenuation Level*/
	signed   int volume;	/*operator attenuation level*/
	unsigned int delta_D1R;	/*volume delta for decay phase*/
	unsigned int D1L;		/*EG switches to D2R, when envelope reaches this level*/
	unsigned int delta_D2R;	/*volume delta for sustain phase*/
	unsigned int delta_RR;	/*volume delta for release phase*/

	unsigned int key;		/*0=last key was KEY OFF, 1=last key was KEY ON*/

	unsigned int KS;		/*Key Scale     */
	unsigned int AR;		/*Attack rate   */
	unsigned int D1R;		/*Decay rate    */
	unsigned int D2R;		/*Sustain rate  */
	unsigned int RR;		/*Release rate  */

	unsigned int reserved0;	/**/
	signed   int reserved1;	/**/
	signed   int reserved2;	/**/

} OscilRec;


typedef struct
{
	OscilRec Oscils[32];	/*there are 32 operators in YM2151*/

	unsigned int PAN[16];	/*channels output masks (0xffffffff = enable)*/

	unsigned int LFOphase;	/*accumulated LFO phase         */
	unsigned int LFOfrq;	/*LFO frequency                 */
	unsigned int LFOwave;	/*LFO waveform (0-saw, 1-square, 2-triangle, 3-random noise)*/
	unsigned int PMD;		/*LFO Phase Modulation Depth    */
	unsigned int AMD;		/*LFO Amplitude Modulation Depth*/
	unsigned int LFA;		/*current AM from LFO*/
	signed   int LFP;		/*current PM from LFO*/

	unsigned int test;		/*TEST register*/

	unsigned int CT;		/*output control pins (bit7 CT2, bit6 CT1)*/

	unsigned int noise;		/*noise register (bit 7 - noise enable, bits 4-0 - noise period*/
	unsigned int noiseRNG;	/*17 bit noise shift register*/
	unsigned int noise_p;	/*noise 'phase'*/
	unsigned int noise_f;	/*noise period*/

	unsigned int CSMreq;	/*CSM KEYON/KEYOFF sequence request*/

	unsigned int IRQenable;	/*IRQ enable for timer B (bit 3) and timer A (bit 2); bit 7 - CSM mode (keyon to all slots, everytime timer A overflows)*/
	unsigned int status;	/*chip status (BUSY, IRQ Flags)*/

///#ifdef USE_MAME_TIMERS
///	void *TimATimer,*TimBTimer;	/*ASG 980324 -- added for tracking timers*/
///	double TimerATime[1024];	/*Timer A times for MAME*/
///	double TimerBTime[256];		/*Timer B times for MAME*/
///#else
	int TimA,TimB;				/*timer A,B enable (0-disabled)*/
	signed int TimAVal,TimBVal;	/*current value of timer*/
	unsigned int TimerA[1024];	/*Timer A deltas*/
	unsigned int TimerB[256];	/*Timer B deltas*/
///#endif
	unsigned int TimAIndex;		/*Timer A index*/
	unsigned int TimBIndex;		/*Timer B index*/

	unsigned int TimAOldIndex;	/*Timer A previous index*/
	unsigned int TimBOldIndex;	/*Timer B previous index*/

	/*
	*   Frequency-deltas to get the closest frequency possible.
	*   There're 11 octaves because of DT2 (max 950 cents over base frequency)
	*   and LFO phase modulation (max 800 cents below AND over base frequency)
	*   Summary:   octave  explanation
	*              0       note code - LFO PM
	*              1       note code
	*              2       note code
	*              3       note code
	*              4       note code
	*              5       note code
	*              6       note code
	*              7       note code
	*              8       note code
	*              9       note code + DT2 + LFO PM
	*              10      note code + DT2 + LFO PM
	*/
	unsigned int freq[11*768];/*11 octaves, 768 'cents' per octave*/

	/*
	*   Frequency deltas for DT1. These deltas alter operator frequency
	*   after it has been taken from frequency-deltas table.
	*/
	signed   int DT1freq[8*32];		/*8 DT1 levels, 32 KC values*/
	unsigned int EG_tab [32+64+32];	/*Envelope Generator deltas (32 + 64 rates + 32 RKS)*/
	unsigned int LFOfreq[256];		/*LFO frequency deltas*/
	unsigned int noise_tab[32];		/*17bit Noise Generator periods*/

///	void (*irqhandler)(int irq);	/*IRQ function handler*/
///	mem_write_handler porthandler;	/*port write function handler*/

	unsigned int clock;				/*chip clock in Hz (passed from 2151intf.c)*/
	unsigned int sampfreq;			/*sampling frequency in Hz (passed from 2151intf.c)*/

} YM2151;


/*
**  Shifts below are subject to change when sampling frequency changes...
*/
#define FREQ_SH			16  /* 16.16 fixed point (frequency calculations) */
#define ENV_SH			16  /* 16.16 fixed point (envelope calculations)  */
#define LFO_SH			23  /*  9.23 fixed point (LFO calculations)       */
#define TIMER_SH		16  /* 16.16 fixed point (timers calculations)    */

#define FREQ_MASK		((1<<FREQ_SH)-1)
#define ENV_MASK		((1<<ENV_SH)-1)

#define ENV_BITS		10
#define ENV_LEN			(1<<ENV_BITS)
#define ENV_STEP		(128.0/ENV_LEN)
#define ENV_QUIET		((int)(0x68/(ENV_STEP)))

#define MAX_ATT_INDEX	((ENV_LEN<<ENV_SH)-1) /*1023.ffff*/
#define MIN_ATT_INDEX	(      (1<<ENV_SH)-1) /*   0.ffff*/

#define EG_ATT			4
#define EG_DEC			3
#define EG_SUS			2
#define EG_REL			1
#define EG_OFF			0

#define SIN_BITS		10
#define SIN_LEN			(1<<SIN_BITS)
#define SIN_MASK		(SIN_LEN-1)

#define TL_RES_LEN		(256) /* 8 bits addressing (real chip) */

#define LFO_BITS		9
#define LFO_LEN			(1<<LFO_BITS)
#define LFO_MASK		(LFO_LEN-1)


///#if (SAMPLE_BITS==16)
	#define FINAL_SH	(0)
	#define MAXOUT		(+32767)
	#define MINOUT		(-32768)
///#else
///	#define FINAL_SH	(8)
///	#define MAXOUT		(+127)
///	#define MINOUT		(-128)
///#endif


/* TL_TAB_LEN is calculated as:
 * 13 - sinus amplitude bits  (Y axis)
 * 2  - sinus sign bit        (Y axis)
 * ENV_LEN - sinus resolution (X axis)
*/
#define TL_TAB_LEN (13*2*TL_RES_LEN)
extern signed int TL_TAB[TL_TAB_LEN];

/* sin waveform table in 'decibel' scale*/
extern unsigned int sin_tab[SIN_LEN];

/* four AM/PM LFO waveforms (8 in total)*/
extern unsigned int lfo_tab[LFO_LEN*4*2];

/* LFO amplitude modulation depth table (128 levels)*/
extern unsigned int lfo_md_tab[128];

/* translate from D1L to volume index (16 D1L levels)*/
extern unsigned int D1L_tab[16];

/*
 *   DT2 defines offset in cents from base note
 *
 *   This table defines offset in frequency-deltas table.
 *   User's Manual page 22
 *
 *   Values below were calculated using formula: value =  orig.val / 1.5625
 *
 *	DT2=0 DT2=1 DT2=2 DT2=3
 *	0     600   781   950
*/
extern unsigned int DT2_tab[4];

/*
 *   DT1 defines offset in Hertz from base note
 *   This table is converted while initialization...
 *   Detune table in YM2151 User's Manual is wrong (checked against the real chip)
*/

extern unsigned char DT1_tab[4*32]; /* 4*32 DT1 values */

extern unsigned short phaseinc_rom[768];




extern YM2151/** * **/YMPSG;	/* array of YM2151's */
///extern unsigned int YMNumChips;	/* total # of YM2151's emulated */


/*these variables stay here because of speedup purposes only */
///extern YM2151 * PSG;
extern signed int chanout[8];
extern signed int c1,m2,c2; /*Phase Modulation input for operators 2,3,4*/



/*save output as raw 16-bit sample*/
/*#define SAVE_SAMPLE*/
/*#define SAVE_SEPARATE_CHANNELS*/

///#if defined SAVE_SAMPLE || defined SAVE_SEPARATE_CHANNELS
///static FILE *sample[9];
///#endif



/* own PI definition */
#ifdef PI
	#undef PI
#endif
#define PI 3.14159265358979323846


#endif /*_H_YM2151_*/
